"""."""

from kinematic_tracker.association.association_metric import AssociationMetric


class MetricMode:
    def __init__(self) -> None:
        self.choices = [m for m in AssociationMetric if m != AssociationMetric.UNKNOWN_METRIC]
        self.metric = self.choices[0]

    def next(self) -> None:
        i = (self.choices.index(self.metric) + 1) % len(self.choices)
        self.metric = self.choices[i]



"""."""

import numpy as np
import pytest

from nuscenes.utils.data_classes import Box
from pyquaternion import Quaternion

from association_studio.boxes import fill_vec_z, from_vec_z


def test_from_vec_z() -> None:
    box = from_vec_z(np.linspace(1, 10, 10))
    assert isinstance(box, Box)
    assert box.center == pytest.approx(np.linspace(1, 3, 3))
    assert box.orientation.q == pytest.approx(np.linspace(4, 7, 4))
    assert box.wlh == pytest.approx(np.linspace(8, 10, 3))


def test_fill_vec_z() -> None:
    q = Quaternion([1.0, 0.1, 0.2, 0.5])
    box = Box([2.0, 3.0, 4.0], [2.0, 4.0, 6.0], q)
    buf_z = np.zeros(10)
    fill_vec_z(box, buf_z)
    assert buf_z == pytest.approx([2, 3, 4, 1.0, 0.1, 0.2, 0.5, 2.0, 4.0, 6.0])



"""."""

import numpy as np

from nuscenes.utils.data_classes import Box
from pyquaternion import Quaternion

from association_studio.common import FVT


def from_vec_z(vec_z: np.ndarray) -> Box:
    return Box(center=vec_z[:3], orientation=Quaternion(vec_z[3:7]), size=vec_z[7:10])


def fill_vec_z(box: Box, vec_z: FVT) -> None:
    vec_z[:3] = box.center
    vec_z[3:7] = box.orientation.q
    vec_z[7:] = box.wlh


def rotate_by_quat(q_rotate: Quaternion, q_io: Quaternion) -> None:
    """Rotate by quaternion.

    Note:
        The operation preserves the positivity of the scalar part of the quaternion.

    Args:
        q_rotate: quaternion representing the rotation.
        q_io: quaternion to be rotated.
    """
    q = q_io * q_rotate
    q = -q if q.q[0] < 0.0 else q
    q_io.q[:] = q.q[:]



"""."""

import numpy as np

from kinematic_tracker.association.association_metric import AssociationMetric
from nuscenes.utils.data_classes import Box

from association_studio.boxes import fill_vec_z
from association_studio.score_driver.tracker_factory import get_low_order_nu_scenes_tracker


class ScoreDriver:
    def __init__(self):
        self.det_1z = np.array([(0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 2.0, 4.0, 1.0)])
        self.buf_1z = np.zeros((1, 10))
        tracker = get_low_order_nu_scenes_tracker()
        tracker.advance(0, self.det_1z)
        self.ref_kalman_filters = [tracker.tracks_c[0][0].kkf.kalman_filter]
        valid_metrics = [
            mt.value for mt in AssociationMetric if mt != AssociationMetric.UNKNOWN_METRIC
        ]
        self.metric_names = np.array(valid_metrics, dtype=str)
        num_metrics = len(self.metric_names)
        self.metric_drivers: np.ndarray[tuple[int], np.dtype[np.object_]] = np.zeros(
            num_metrics, object
        )
        self.scores = np.zeros(num_metrics)
        self._init_metric_drivers()
        self.metric_mask = np.zeros(num_metrics, bool)
        self.metric_index = -1
        self.score = 0.0
        self.set_association_metric(AssociationMetric.SIZE_MODULATED_MAHALANOBIS.value)

    def _init_metric_drivers(self) -> None:
        tracker = get_low_order_nu_scenes_tracker()
        for metric_num, metric_name in enumerate(self.metric_names):
            tracker.set_association_metric(metric_name)
            self.metric_drivers[metric_num] = tracker.metric_driver

    def set_association_metric(self, metric: str) -> None:
        self.metric_mask[:] = metric == self.metric_names
        self.metric_index = np.argmax(self.metric_mask)

    def compute_score(self, box: Box) -> None:
        self.compute_scores(box)
        assert self.metric_index >= 0
        self.score = self.scores[self.metric_index]

    def compute_scores(self, box: Box) -> None:
        fill_vec_z(box, self.buf_1z[0])
        for idx, driver in enumerate(self.metric_drivers):
            self.scores[idx] = driver.compute_metric(self.buf_1z, self.ref_kalman_filters)[0, 0]



"""."""

import pytest

from pyquaternion import Quaternion

from association_studio.boxes import rotate_by_quat


def test_rotate1() -> None:
    q_rot = Quaternion([1.0, 0.0, 0.0, 0.5])
    q_io = Quaternion([1.0, 0.0, 0.0, 0.5])
    rotate_by_quat(q_rot, q_io)
    assert q_io.q == pytest.approx([0.75, 0.0, 0.0, 1.0])


def test_rotate2() -> None:
    q_rot = Quaternion([-1.0, 0.0, 0.0, 1.0])
    q_io = Quaternion([1.0, 0.0, 0.0, 0.0])
    rotate_by_quat(q_rot, q_io)
    assert q_io.q == pytest.approx([1.0, 0.0, 0.0, -1.0])



"""."""

import copy

import matplotlib.pyplot as plt
import numpy as np

from matplotlib.backend_bases import MouseEvent, PickEvent
from matplotlib.container import BarContainer
from matplotlib.patches import Circle, Polygon
from pyquaternion import Quaternion
from scipy.spatial.transform import Rotation

from association_studio.boxes import from_vec_z, rotate_by_quat
from association_studio.score_driver.metric_labels import get_labels
from association_studio.score_driver.score_driver import ScoreDriver

from .metric_mode import MetricMode
from .show_coo_mode import ShowCooMode


class AssociationStudioBackend:
    def __init__(self) -> None:
        self.score_driver = ScoreDriver()
        self.show_coo_mode = ShowCooMode.TRANSLATION_AND_ROTATION
        self.metric_mode = MetricMode()
        self.box_det = from_vec_z(self.score_driver.det_1z[0])
        self.box_trk = from_vec_z(np.array((0.3, 0.7, 0.0, 1.0, 0.0, 0.0, -0.4, 2.0, 4.0, 1.0)))
        self.box_trk_start_pose_change = copy.deepcopy(self.box_trk)
        bcc = self.box_det.bottom_corners()[:2, :].T
        self.det_poly = Polygon(bcc, edgecolor='k', facecolor='yellow', closed=True)
        self.det_circle = Circle(bcc[2:].mean(axis=0), 0.35, edgecolor='k', facecolor='yellow')

        bcc = self.box_trk.bottom_corners()[:2, :].T
        self.trk_poly = Polygon(bcc, edgecolor='b', facecolor='none', closed=True, picker=True)
        trk_xy = bcc[2:].mean(axis=0)
        self.trk_circle = Circle(trk_xy, 0.35, edgecolor='b', facecolor='w', picker=True)
        self.score_driver.compute_score(self.box_trk)

    def next_show_coo_mode(self) -> None:
        self.show_coo_mode = ShowCooMode(self.show_coo_mode.next())

    def next_metric(self) -> None:
        self.metric_mode.next()
        self.score_driver.set_association_metric(self.metric_mode.metric.value)
        self.score_driver.compute_score(self.box_trk)

    def start_pose_change(self) -> None:
        self.box_trk_start_pose_change = copy.deepcopy(self.box_trk)

    def change_trk_pose(self, pick: PickEvent, event: MouseEvent) -> None:
        me = pick.mouseevent

        if pick.artist == self.trk_poly:
            translation = np.array((event.xdata - me.xdata, event.ydata - me.ydata, 0.0))
            self.box_trk = copy.deepcopy(self.box_trk_start_pose_change)
            self.box_trk.translate(translation)
        elif pick.artist == self.trk_circle:
            self.box_trk = copy.deepcopy(self.box_trk_start_pose_change)
            c = self.box_trk.center
            a = np.array((event.xdata - c[0], event.ydata - c[1], 0.0))
            b = np.array((me.xdata - c[0], me.ydata - c[1], 0.0))
            q_ab = Quaternion(Rotation.align_vectors(a, b)[0].as_quat(scalar_first=True))
            rotate_by_quat(q_ab, self.box_trk.orientation)

        bcc = self.box_trk.bottom_corners()[:2, :].T
        self.trk_poly.set_xy(bcc)
        self.trk_circle.set_center(bcc[2:].mean(axis=0))
        self.score_driver.compute_score(self.box_trk)

    def get_title(self) -> str:
        return f'{self.metric_mode.metric.value} {self.score_driver.score:5.4f}'

    def get_text(self) -> str:
        return self.show_coo_mode.format(self.box_trk)

    def get_horizontal_bar_container(self, ax_metrics: plt.Axes) -> BarContainer:
        scores = self.score_driver.scores
        names = get_labels(self.score_driver.metric_names)
        return ax_metrics.barh(range(scores.shape[0]), scores, tick_label=names)

    def update_horizontal_bar_container(self, bar: BarContainer) -> None:
        for bar, new_h in zip(bar, self.score_driver.scores):
            bar.set_width(new_h)



"""."""

from kinematic_tracker.association.association_metric import AssociationMetric
from matplotlib.backend_bases import KeyEvent

from association_studio.app.front import AssociationStudioFrontend
from association_studio.app.show_coo_mode import ShowCooMode


def test_coo_mode(front: AssociationStudioFrontend, key_event: KeyEvent) -> None:
    front.on_key(key_event)
    assert front.back.show_coo_mode == ShowCooMode.TRANSLATION
    assert front.text.get_text() == 'xy 0.300 0.700'


def test_metric(front: AssociationStudioFrontend, key_event: KeyEvent) -> None:
    key_event.key = 'm'
    front.on_key(key_event)
    assert front.back.metric_mode.metric == AssociationMetric.MAHALANOBIS
    assert front.ax_boxes.get_title() == 'mahalanobis 0.9821'



"""."""

import pytest

from nuscenes.utils.data_classes import Box

from association_studio.score_driver.score_driver import ScoreDriver


def test_compute_scores(box: Box, driver: ScoreDriver) -> None:
    driver.compute_scores(box)
    ref = [0.13737373737373737, 0.1866537467790667, 0.7054992044360596, 0.5343434343434343]
    assert driver.scores == pytest.approx(ref)
